@@ -35,8 +35,6 @@
 #include "utils.h"
 
 
-#define THE_ANSWER_TO_THE_ULTIMATE_QUESTION_OF_LIFE_THE_UNIVERSE_AND_EVERYTHING 42
-
 #define REDSOCKS_RELAY_HALFBUFF  4096
 
 
@@ -66,9 +64,38 @@ static parser_entry redsocks_entries[] =
 	{ .key = "type",       .type = pt_pchar },
 	{ .key = "login",      .type = pt_pchar },
 	{ .key = "password",   .type = pt_pchar },
+	{ .key = "listenq",    .type = pt_uint16 },
+	{ .key = "min_accept_backoff", .type = pt_uint16 },
+	{ .key = "max_accept_backoff", .type = pt_uint16 },
 	{ }
 };
 
+/* There is no way to get `EVLIST_INSERTED` event flag outside of libevent, so
+ * here are tracking functions. */
+static void tracked_event_set(
+		struct tracked_event *tev, evutil_socket_t fd, short events,
+		void (*callback)(evutil_socket_t, short, void *), void *arg)
+{
+	event_set(&tev->ev, fd, events, callback, arg);
+	timerclear(&tev->inserted);
+}
+
+static int tracked_event_add(struct tracked_event *tev, const struct timeval *tv)
+{
+	int ret = event_add(&tev->ev, tv);
+	if (ret == 0)
+		gettimeofday(&tev->inserted, NULL);
+	return ret;
+}
+
+static int tracked_event_del(struct tracked_event *tev)
+{
+	int ret = event_del(&tev->ev);
+	if (ret == 0)
+		timerclear(&tev->inserted);
+	return ret;
+}
+
 static int redsocks_onenter(parser_section *section)
 {
 	// FIXME: find proper way to calulate instance_payload_len
@@ -90,6 +117,12 @@ static int redsocks_onenter(parser_section *section)
 	instance->config.bindaddr.sin_addr.s_addr = htonl(INADDR_LOOPBACK);
 	instance->config.relayaddr.sin_family = AF_INET;
 	instance->config.relayaddr.sin_addr.s_addr = htonl(INADDR_LOOPBACK);
+	/* Default value can be checked in run-time, but I doubt anyone needs that.
+	 * Linux:   sysctl net.core.somaxconn
+	 * FreeBSD: sysctl kern.ipc.somaxconn */
+	instance->config.listenq = SOMAXCONN;
+	instance->config.min_backoff_ms = 100;
+	instance->config.max_backoff_ms = 60000;
 
 	for (parser_entry *entry = &section->entries[0]; entry->key; entry++)
 		entry->addr =
@@ -100,6 +133,9 @@ static int redsocks_onenter(parser_section *section)
 			(strcmp(entry->key, "type") == 0)       ? (void*)&instance->config.type :
 			(strcmp(entry->key, "login") == 0)      ? (void*)&instance->config.login :
 			(strcmp(entry->key, "password") == 0)   ? (void*)&instance->config.password :
+			(strcmp(entry->key, "listenq") == 0)    ? (void*)&instance->config.listenq :
+			(strcmp(entry->key, "min_accept_backoff") == 0) ? (void*)&instance->config.min_backoff_ms :
+			(strcmp(entry->key, "max_accept_backoff") == 0) ? (void*)&instance->config.max_backoff_ms :
 			NULL;
 	section->data = instance;
 	return 0;
@@ -137,6 +173,18 @@ static int redsocks_onexit(parser_section *section)
 		err = "no `type` for redsocks";
 	}
 
+	if (!err && !instance->config.min_backoff_ms) {
+		err = "`min_accept_backoff` must be positive, 0 ms is too low";
+	}
+
+	if (!err && !instance->config.max_backoff_ms) {
+		err = "`max_accept_backoff` must be positive, 0 ms is too low";
+	}
+
+	if (!err && !(instance->config.min_backoff_ms < instance->config.max_backoff_ms)) {
+		err = "`min_accept_backoff` must be less than `max_accept_backoff`";
+	}
+
 	if (err)
 		parser_error(section->context, err);
 
@@ -292,12 +340,12 @@ void redsocks_drop_client(redsocks_client *client)
 		client->instance->relay_ss->fini(client);
 
 	if (client->client) {
-		close(EVENT_FD(&client->client->ev_write));
+		redsocks_close(EVENT_FD(&client->client->ev_write));
 		bufferevent_free(client->client);
 	}
 
 	if (client->relay) {
-		close(EVENT_FD(&client->relay->ev_write));
+		redsocks_close(EVENT_FD(&client->relay->ev_write));
 		bufferevent_free(client->relay);
 	}
 
@@ -531,6 +579,45 @@ void redsocks_connect_relay(redsocks_client *client)
 	}
 }
 
+static void redsocks_accept_backoff(int fd, short what, void *_arg)
+{
+	redsocks_instance *self = _arg;
+
+	/* Isn't it already deleted? EV_PERSIST has nothing common with timeouts in
+	 * old libevent... On the other hand libevent does not return any error. */
+	if (tracked_event_del(&self->accept_backoff) != 0)
+		log_errno(LOG_ERR, "event_del");
+
+	if (tracked_event_add(&self->listener, NULL) != 0)
+		log_errno(LOG_ERR, "event_add");
+}
+
+void redsocks_close_internal(int fd, const char* file, int line, const char *func)
+{
+	if (close(fd) == 0) {
+		redsocks_instance *instance = NULL;
+		struct timeval now;
+		gettimeofday(&now, NULL);
+		list_for_each_entry(instance, &instances, list) {
+			if (timerisset(&instance->accept_backoff.inserted)) {
+				struct timeval min_accept_backoff = {
+					instance->config.min_backoff_ms / 1000,
+					(instance->config.min_backoff_ms % 1000) * 1000};
+				struct timeval time_passed;
+				timersub(&now, &instance->accept_backoff.inserted, &time_passed);
+				if (timercmp(&min_accept_backoff, &time_passed, <)) {
+					redsocks_accept_backoff(-1, 0, instance);
+					break;
+				}
+			}
+		}
+	}
+	else {
+		const int do_errno = 1;
+		_log_write(file, line, func, do_errno, LOG_WARNING, "close");
+	}
+}
+
 static void redsocks_accept_client(int fd, short what, void *_arg)
 {
 	redsocks_instance *self = _arg;
@@ -546,9 +633,25 @@ static void redsocks_accept_client(int fd, short what, void *_arg)
 	// working with client_fd
 	client_fd = accept(fd, (struct sockaddr*)&clientaddr, &addrlen);
 	if (client_fd == -1) {
-		log_errno(LOG_WARNING, "accept");
+		/* Different systems use different `errno` value to signal different
+		 * `lack of file descriptors` conditions. Here are most of them.  */
+		if (errno == ENFILE || errno == EMFILE || errno == ENOBUFS || errno == ENOMEM) {
+			self->accept_backoff_ms = (self->accept_backoff_ms << 1) + 1;
+			clamp_value(self->accept_backoff_ms, self->config.min_backoff_ms, self->config.max_backoff_ms);
+			int delay = (random() % self->accept_backoff_ms) + 1;
+			log_errno(LOG_WARNING, "accept: out of file descriptors, backing off for %u ms", delay);
+			struct timeval tvdelay = { delay / 1000, (delay % 1000) * 1000 };
+			if (tracked_event_del(&self->listener) != 0)
+				log_errno(LOG_ERR, "event_del");
+			if (tracked_event_add(&self->accept_backoff, &tvdelay) != 0)
+				log_errno(LOG_ERR, "event_add");
+		}
+		else {
+			log_errno(LOG_WARNING, "accept");
+		}
 		goto fail;
 	}
+	self->accept_backoff_ms = 0;
 
 	// socket is really bound now (it could be bound to 0.0.0.0)
 	addrlen = sizeof(myaddr);
@@ -615,7 +718,7 @@ fail:
 		redsocks_drop_client(client);
 	}
 	if (client_fd != -1)
-		close(client_fd);
+		redsocks_close(client_fd);
 }
 
 static const char *redsocks_evshut_str(unsigned short evshut)
@@ -705,14 +808,18 @@ static int redsocks_init_instance(redsocks_instance *instance)
 		goto fail;
 	}
 
-	error = listen(fd, THE_ANSWER_TO_THE_ULTIMATE_QUESTION_OF_LIFE_THE_UNIVERSE_AND_EVERYTHING); // does anyone know better value?
+	error = listen(fd, instance->config.listenq);
 	if (error) {
 		log_errno(LOG_ERR, "listen");
 		goto fail;
 	}
 
-	event_set(&instance->listener, fd, EV_READ | EV_PERSIST, redsocks_accept_client, instance);
-	error = event_add(&instance->listener, NULL);
+	tracked_event_set(&instance->listener, fd, EV_READ | EV_PERSIST, redsocks_accept_client, instance);
+	fd = -1;
+
+	tracked_event_set(&instance->accept_backoff, -1, 0, redsocks_accept_backoff, instance);
+
+	error = tracked_event_add(&instance->listener, NULL);
 	if (error) {
 		log_errno(LOG_ERR, "event_add");
 		goto fail;
@@ -724,8 +831,7 @@ fail:
 	redsocks_fini_instance(instance);
 
 	if (fd != -1) {
-		if (close(fd) != 0)
-			log_errno(LOG_WARNING, "close");
+		redsocks_close(fd);
 	}
 
 	return -1;
@@ -747,14 +853,21 @@ static void redsocks_fini_instance(redsocks_instance *instance) {
 	if (instance->relay_ss->instance_fini)
 		instance->relay_ss->instance_fini(instance);
 
-	if (event_initialized(&instance->listener)) {
-		if (event_del(&instance->listener) != 0)
-			log_errno(LOG_WARNING, "event_del");
-		if (close(EVENT_FD(&instance->listener)) != 0)
-			log_errno(LOG_WARNING, "close");
+	if (event_initialized(&instance->listener.ev)) {
+		if (timerisset(&instance->listener.inserted))
+			if (tracked_event_del(&instance->listener) != 0)
+				log_errno(LOG_WARNING, "event_del");
+		redsocks_close(EVENT_FD(&instance->listener.ev));
 		memset(&instance->listener, 0, sizeof(instance->listener));
 	}
 
+	if (event_initialized(&instance->accept_backoff.ev)) {
+		if (timerisset(&instance->accept_backoff.inserted))
+			if (tracked_event_del(&instance->accept_backoff) != 0)
+				log_errno(LOG_WARNING, "event_del");
+		memset(&instance->accept_backoff, 0, sizeof(instance->accept_backoff));
+	}
+
 	list_del(&instance->list);
 
 	free(instance->config.type);
